from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor

from .. import CoreService_pb2 as pb
from ..job_info import JobInfo


class BaseJob(ABC):
    def __init__(self, stub, job_id):
        self._stub = stub
        self._job_id = job_id
        self._result = None
        self._job = self._stub.GetJob(self._job_id)
        self._thread_pool = ThreadPoolExecutor(5)
        self._info = JobInfo(self._stub, self._job_id)

    @property
    def id(self):
        return self._job_id.job_id

    @property
    def job_type(self):
        return pb.JobType.Name(self._job.job_type)

    @property
    def done(self):
        return self.is_done()

    @property
    def child_job_ids(self):
        return self._job.childJobIds

    @abstractmethod
    def _response_method(self, job_id):
        raise NotImplementedError(
            "Method `_response_method` needs to be implemented by the child class"
        )

    def is_done(self) -> bool:
        return self._stub.GetJob(self._job_id).done

    def get_result(self):
        if not self._result:
            if not self.is_done():
                raise Exception("Job has not finished yet!")
            self._result = self._response_method(self._job_id)
        return self._result

    def show_progress(self):
        self._info.show_progress()

    def get_metrics(self):
        return self._info.get_metrics()

    def __repr__(self):
        return f"""Job(id={self.id}, type={self.job_type}, done={self.done}, childJobIds={self.child_job_ids})"""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: featurestore/core/RecommendationProtoApi.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='featurestore/core/RecommendationProtoApi.proto',
  package='featurestore.core',
  syntax='proto3',
  serialized_options=b'\n\030ai.h2o.featurestore.coreB\033RecommendationProtoApiProto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n.featurestore/core/RecommendationProtoApi.proto\x12\x11\x66\x65\x61turestore.core\"h\n$ListRecommendationClassifierResponse\x12@\n\x0b\x63lassifiers\x18\x01 \x03(\x0b\x32+.featurestore.core.RecommendationClassifier\"@\n%CreateRecommendationClassifierRequest\x12\x17\n\x0f\x63lassifier_name\x18\x01 \x01(\t\"(\n\x18RecommendationClassifier\x12\x0c\n\x04name\x18\x01 \x01(\tB7\n\x18\x61i.h2o.featurestore.coreB\x1bRecommendationProtoApiProtob\x06proto3'
)




_LISTRECOMMENDATIONCLASSIFIERRESPONSE = _descriptor.Descriptor(
  name='ListRecommendationClassifierResponse',
  full_name='featurestore.core.ListRecommendationClassifierResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='classifiers', full_name='featurestore.core.ListRecommendationClassifierResponse.classifiers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=69,
  serialized_end=173,
)


_CREATERECOMMENDATIONCLASSIFIERREQUEST = _descriptor.Descriptor(
  name='CreateRecommendationClassifierRequest',
  full_name='featurestore.core.CreateRecommendationClassifierRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='classifier_name', full_name='featurestore.core.CreateRecommendationClassifierRequest.classifier_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=175,
  serialized_end=239,
)


_RECOMMENDATIONCLASSIFIER = _descriptor.Descriptor(
  name='RecommendationClassifier',
  full_name='featurestore.core.RecommendationClassifier',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='featurestore.core.RecommendationClassifier.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=241,
  serialized_end=281,
)

_LISTRECOMMENDATIONCLASSIFIERRESPONSE.fields_by_name['classifiers'].message_type = _RECOMMENDATIONCLASSIFIER
DESCRIPTOR.message_types_by_name['ListRecommendationClassifierResponse'] = _LISTRECOMMENDATIONCLASSIFIERRESPONSE
DESCRIPTOR.message_types_by_name['CreateRecommendationClassifierRequest'] = _CREATERECOMMENDATIONCLASSIFIERREQUEST
DESCRIPTOR.message_types_by_name['RecommendationClassifier'] = _RECOMMENDATIONCLASSIFIER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ListRecommendationClassifierResponse = _reflection.GeneratedProtocolMessageType('ListRecommendationClassifierResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRECOMMENDATIONCLASSIFIERRESPONSE,
  '__module__' : 'featurestore.core.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.ListRecommendationClassifierResponse)
  })
_sym_db.RegisterMessage(ListRecommendationClassifierResponse)

CreateRecommendationClassifierRequest = _reflection.GeneratedProtocolMessageType('CreateRecommendationClassifierRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATERECOMMENDATIONCLASSIFIERREQUEST,
  '__module__' : 'featurestore.core.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.CreateRecommendationClassifierRequest)
  })
_sym_db.RegisterMessage(CreateRecommendationClassifierRequest)

RecommendationClassifier = _reflection.GeneratedProtocolMessageType('RecommendationClassifier', (_message.Message,), {
  'DESCRIPTOR' : _RECOMMENDATIONCLASSIFIER,
  '__module__' : 'featurestore.core.RecommendationProtoApi_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.RecommendationClassifier)
  })
_sym_db.RegisterMessage(RecommendationClassifier)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



class Recommendation:
    def __init__(self, stub, source_feature_set, match_result):
        from .feature_set import FeatureSet  # To avoid circular import reference

        self._source = source_feature_set
        self._target = FeatureSet(stub, match_result.target_feature_set)
        self._joins = {join.classifier: join.columns for join in match_result.joins}

    def __repr__(self):
        return str(
            {
                "source": {
                    "project_name": self._source.project,
                    "feature_set_name": self._source.feature_set_name,
                },
                "target": {
                    "project_name": self._target.project,
                    "feature_set_name": self._target.feature_set_name,
                },
                "joins": self._joins,
            }
        )



from typing import List

from .. import CoreService_pb2 as pb
from ..entities.base_job import BaseJob
from ..entities.compute_statistics_job import ComputeStatisticsJob
from ..entities.extract_schema_job import ExtractSchemaJob
from ..entities.ingest_job import IngestJob
from ..entities.materialization_online_job import MaterializationOnlineJob
from ..entities.retrieve_job import RetrieveJob
from ..entities.revert_ingest_job import RevertIngestJob


class Jobs:
    def __init__(self, stub):
        self._stub = stub

    @staticmethod
    def _create_job(stub, job_proto):
        job_id = pb.JobId(job_id=job_proto.job_id)
        if job_proto.job_type == pb.JobType.Ingest:
            return IngestJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.ExtractSchema:
            return ExtractSchemaJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.Retrieve:
            return RetrieveJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.MaterializationOnline:
            return MaterializationOnlineJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.ComputeStatistics:
            return ComputeStatisticsJob(stub, job_id)
        elif job_proto.job_type == pb.JobType.Revert:
            return RevertIngestJob(stub, job_id)

    def list(self, active=True, job_type=pb.JobType.Unknown) -> List[BaseJob]:
        request = pb.ListJobsRequest(active=active, job_type=job_type)
        resp = self._stub.ListJobs(request)
        return [Jobs._create_job(self._stub, job_proto) for job_proto in resp.jobs]

    def get(self, job_id: str) -> BaseJob:
        request = pb.JobId(job_id=job_id)
        job_proto = self._stub.GetJob(request)
        return Jobs._create_job(self._stub, job_proto)

    def __repr__(self):
        return "This class wraps together methods working with jobs"



import featurestore.core.CoreService_pb2 as pb

from . import interactive_console
from .config import ConfigUtils
from .utils import Utils


class JobInfo:
    def __init__(self, stub, job_id):
        self._stub = stub
        self._job_id = job_id
        self._next_index = 0

    def show_progress(self):
        if ConfigUtils.is_interactive_print_enabled():
            request = pb.JobProgressInput(
                job_id=self._job_id.job_id, next_index=self._next_index
            )
            response = self._stub.GetJobProgress(request)
            if response.progress:
                self._next_index += len(response.progress)
                for progress in response.progress:
                    interactive_console.log(
                        f"Job ID: {self._job_id.job_id}, Status: {progress.message}"
                    )

    def get_metrics(self):
        if self._stub.GetJob(self._job_id).done:
            request = pb.JobProgressInput(job_id=self._job_id.job_id, next_index=0)
            response = self._stub.GetJobProgress(request)
            if response.progress:
                return {
                    progress.message: f"{progress.duration_in_seconds}s"
                    for progress in response.progress
                }
        else:
            interactive_console.log(
                "Job is still running. Please wait until the job finishes."
            )



import os
import tempfile
import time
from concurrent.futures import ThreadPoolExecutor

import requests

from . import CoreService_pb2 as pb
from .commons.spark_utils import SparkUtils
from .job_info import JobInfo


class RetrieveAsLinksCommon:
    def _get_job(self, job_id):
        return self._stub.GetJob(job_id)

    def _get_retrieve_links_output(self, job_id):
        return self._stub.GetRetrieveAsLinksJobOutput(job_id)

    def download(self, output_dir=None):
        info = JobInfo(self._stub, self._job_id)
        while not self._get_job(self._job_id).done:
            info.show_progress()
            time.sleep(2)
        info.show_progress()  # there is possibility that some progress was pushed before finishing job
        retrieve_as_links_response = self._get_retrieve_links_output(self._job_id)
        return RetrieveAsLinksCommon._download_all(
            output_dir, retrieve_as_links_response.download_links
        )

    def download_async(self, output_dir=None):
        future = self._thread_pool.submit(self.download, output_dir)
        return DownloadFuture(future)

    @staticmethod
    def _download_all(output_dir, download_links):
        if output_dir is None:
            output_dir = tempfile.mkdtemp()
        for idx, path in list(enumerate(download_links)):
            dest = os.path.join(output_dir, "part_" + str(idx) + ".parquet")
            RetrieveAsLinksCommon._download_single(dest, path)
        return output_dir

    @staticmethod
    def _download_single(dest, origin):
        with requests.get(origin, stream=True) as r:
            r.raise_for_status()
            with open(dest, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        return dest


class RetrieveHolder(RetrieveAsLinksCommon):
    def __init__(self, stub, feature_set, start_date_time, end_date_time, ingest_id):
        self._stub = stub
        self._thread_pool = ThreadPoolExecutor(5)
        self._feature_set = feature_set
        self._start_date_time = start_date_time
        self._end_date_time = end_date_time
        self._retrieve_as_spark_response = None
        self._job_id = None
        self._ingest_id = ingest_id

    def _start_retrieve_links_job(self):
        request = self._create_retrieve_request()
        return self._stub.StartRetrieveAsLinksJob(request)

    def download(self, output_dir=None):
        if not self._job_id:
            self._job_id = self._start_retrieve_links_job()
        return super(RetrieveHolder, self).download(output_dir)

    def download_async(self, output_dir=None):
        if not self._job_id:
            self._job_id = self._start_retrieve_links_job()
        return super(RetrieveHolder, self).download_async(output_dir)

    def _start_preview_job(self, num_rows):
        request = pb.PreviewRequest()
        request.feature_set.CopyFrom(self._feature_set)
        if self._start_date_time is not None:
            request.start_date_time = self._start_date_time
        if self._end_date_time is not None:
            request.end_date_time = self._end_date_time
        request.num_rows = num_rows
        request.ingest_id = self._ingest_id
        return self._stub.StartPreviewJob(request)

    def preview(self, num_rows=10):
        if num_rows > 100:
            print(
                "Preview can retrieve at most 100 rows. Please try with a lesser value."
            )
            return

        if not self._job_id:
            self._job_id = self._start_preview_job(num_rows)

        info = JobInfo(self._stub, self._job_id)
        while not self._get_job(self._job_id).done:
            info.show_progress()
            time.sleep(2)
        info.show_progress()  # there is possibility that some progress was pushed before finishing job
        resp = self._stub.GetPreviewJobOutput(self._job_id)
        print(resp.result)

    def as_spark_frame(self, spark_session):
        from pyspark.sql.functions import (
            col,
            from_utc_timestamp,
            lit,
            to_timestamp,
            unix_timestamp,
        )

        if self._retrieve_as_spark_response is None:
            request = self._create_retrieve_request(spark_session)
            self._retrieve_as_spark_response = self._stub.RetrieveAsSpark(request)
        resp = self._retrieve_as_spark_response
        spark_session.conf.set("ai.h2o.featurestore.sessionId", resp.session_id)
        SparkUtils.configure_user_spark(spark_session)
        for k, v in resp.options.items():
            spark_session.conf.set(k, v)
        df = spark_session.read.format("delta").load(resp.cache_path)
        retrieve_scope = resp.retrieve_scope
        start_timestamp = retrieve_scope.start_date_time.seconds
        end_timestamp = retrieve_scope.end_date_time.seconds
        timestamp_col = "timestamp_" + str(round(time.time() * 1000))
        if self._ingest_id:
            output_df = df.filter(col("ingest_id") == lit(self._ingest_id))
        elif self._feature_set.time_travel_column:
            output_df = (
                df.withColumn(
                    timestamp_col,
                    unix_timestamp(
                        from_utc_timestamp(
                            to_timestamp(
                                col("`" + self._feature_set.time_travel_column + "`"),
                                self._feature_set.time_travel_column_format,
                            ),
                            spark_session.conf.get("spark.sql.session.timeZone"),
                        )
                    ),
                )
                .filter(col(timestamp_col) <= end_timestamp)
                .filter(col(timestamp_col) >= start_timestamp)
            ).drop(timestamp_col)
        else:
            output_df = df.filter(
                col("time_travel_column_auto_generated") <= end_timestamp
            ).filter(col("time_travel_column_auto_generated") >= start_timestamp)
        internal_columns = ["ingest_id"] + [
            column for column in df.columns if column.startswith("__")
        ]
        return output_df.drop(*internal_columns)

    def _create_retrieve_request(self, spark_session=None):
        session_id = ""
        if spark_session is not None:
            session_id = spark_session.conf.get("ai.h2o.featurestore.sessionId", "")
        request = pb.RetrieveRequest()
        request.feature_set.CopyFrom(self._feature_set)
        request.session_id = session_id
        if self._start_date_time is not None:
            request.start_date_time = self._start_date_time
        if self._end_date_time is not None:
            request.end_date_time = self._end_date_time
        request.ingest_id = self._ingest_id
        return request


class DownloadFuture:
    def __init__(self, future):
        self._future = future
        self._result = None

    def is_done(self) -> bool:
        return self._future.done()

    def get_result(self):
        if not self._result:
            if not self.is_done():
                raise Exception("Job has not finished yet!")
            self._result = self._future.result()
        return self._result



from . import CoreService_pb2 as pb

TINYINT = pb.FeatureDataType.Name(pb.FeatureDataType.TinyInt).upper()
SMALLINT = pb.FeatureDataType.Name(pb.FeatureDataType.SmallInt).upper()
BIGINT = pb.FeatureDataType.Name(pb.FeatureDataType.BigInt).upper()
INT = pb.FeatureDataType.Name(pb.FeatureDataType.Int).upper()
DOUBLE = pb.FeatureDataType.Name(pb.FeatureDataType.Double).upper()
FLOAT = pb.FeatureDataType.Name(pb.FeatureDataType.Float).upper()
STRING = pb.FeatureDataType.Name(pb.FeatureDataType.String).upper()
BINARY = pb.FeatureDataType.Name(pb.FeatureDataType.Binary).upper()
BOOLEAN = pb.FeatureDataType.Name(pb.FeatureDataType.Boolean).upper()
DATE = pb.FeatureDataType.Name(pb.FeatureDataType.Date).upper()
TIMESTAMP = pb.FeatureDataType.Name(pb.FeatureDataType.Timestamp).upper()



from . import CoreService_pb2 as pb

INGEST = pb.JobType.Ingest
RETRIEVE = pb.JobType.Retrieve
EXTRACT_SCHEMA = pb.JobType.ExtractSchema
REVERT_INGEST = pb.JobType.Revert
MATERIALIZATION_ONLINE = pb.JobType.MaterializationOnline
COMPUTE_STATISTICS = pb.JobType.ComputeStatistics



from typing import List

from google.protobuf.empty_pb2 import Empty

from .. import CoreService_pb2 as pb
from ..entities.project import Project


class Projects:
    def __init__(self, stub):
        self._stub = stub
        self._default_lock_value = None

    def list(self) -> List[Project]:
        request = Empty()
        response = self._stub.ListProjects(request)
        return [Project(self._stub, p) for p in response.projects]

    def create(
        self,
        project_name: str,
        description: str = "",
        secret: bool = False,
        locked: bool = None,
    ) -> Project:

        if locked is None and self._default_lock_value is None:
            self._default_lock_value = self._stub.GetProjectsDefault(Empty()).locked
        if locked is None:
            locked_value = self._default_lock_value
        else:
            locked_value = locked
        request = pb.CreateProjectRequest()
        request.secret = secret
        request.project_name = project_name
        request.description = description
        request.locked = locked_value
        response = self._stub.CreateProject(request)
        if response.already_exists:
            print("Project '" + project_name + "' already exists.")
        return Project(self._stub, response.project)

    def get(self, project_name: str) -> Project:
        request = pb.GetProjectRequest()
        request.project_name = project_name
        response = self._stub.GetProject(request)
        return Project(self._stub, response.project)

    def __repr__(self):
        return "This class wraps together methods working with projects"



from google.protobuf.empty_pb2 import Empty

from .. import RecommendationProtoApi_pb2 as RecommendationApi


class Classifiers:
    def __init__(self, stub):
        self._stub = stub

    def list(self) -> [str]:
        request = Empty()
        response = self._stub.ListRecommendationClassifiers(request)
        return [classifier.name for classifier in response.classifiers]

    def create(self, name: str) -> None:
        request = RecommendationApi.CreateRecommendationClassifierRequest()
        request.classifier_name = name
        self._stub.CreateRecommendationClassifier(request)

    def __repr__(self):
        return "Recommendation classifiers"



from .. import CoreService_pb2 as pb
from .. import FeatureSetProtoApi_pb2 as FeatureSetApi
from ..utils import Utils

CATEGORICAL = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.CATEGORICAL)
NUMERICAL = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.NUMERICAL)
TEMPORAL = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.TEMPORAL)
TEXT = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.TEXT)
COMPOSITE = FeatureSetApi.FeatureType.Name(FeatureSetApi.FeatureType.COMPOSITE)
AUTOMATIC_DISCOVERY = FeatureSetApi.FeatureType.Name(
    FeatureSetApi.FeatureType.AUTOMATIC_DISCOVERY
)


class Feature:
    def __init__(self, stub, feature_set, internal_feature, absolute_feature_name):
        self._stub = stub
        self._fs = feature_set
        self._internal_feature = internal_feature
        self._absolute_feature_name = absolute_feature_name

    @property
    def name(self):
        return self._internal_feature.name

    @property
    def version(self):
        return self._internal_feature.version

    @property
    def special(self):
        return self._internal_feature.special

    @special.setter
    def special(self, value):
        update_request = pb.FeatureStringFieldUpdateRequest()
        update_request.new_value = value
        update_request.absolute_feature_name = self._absolute_feature_name
        update_request.header.CopyFrom(self._fs._feature_set_header)
        self._stub.UpdateFeatureSpecial(update_request)
        self._refresh()

    @property
    def version_change(self):
        return self._internal_feature.version_change

    @property
    def status(self):
        return self._internal_feature.status

    @status.setter
    def status(self, value):
        update_request = pb.FeatureStringFieldUpdateRequest()
        update_request.new_value = value
        update_request.absolute_feature_name = self._absolute_feature_name
        update_request.header.CopyFrom(self._fs._feature_set_header)
        self._stub.UpdateFeatureStatus(update_request)
        self._refresh()

    @property
    def data_type(self):
        return self._internal_feature.data_type

    @property
    def profile(self):
        return FeatureProfile(self._stub, self._fs, self)

    @property
    def description(self):
        return self._internal_feature.description

    @description.setter
    def description(self, value):
        update_request = pb.FeatureStringFieldUpdateRequest()
        update_request.new_value = value
        update_request.absolute_feature_name = self._absolute_feature_name
        update_request.header.CopyFrom(self._fs._feature_set_header)
        self._stub.UpdateFeatureDescription(update_request)
        self._refresh()

    @property
    def classifier(self):
        return self._internal_feature.classifier

    @classifier.setter
    def classifier(self, value):
        update_request = pb.FeatureStringFieldUpdateRequest()
        update_request.new_value = value
        update_request.absolute_feature_name = self._absolute_feature_name
        update_request.header.CopyFrom(self._fs._feature_set_header)
        self._stub.UpdateFeatureClassifier(update_request)
        self._refresh()

    @property
    def importance(self):
        return self._internal_feature.importance

    @importance.setter
    def importance(self, value):
        update_request = pb.FeatureDoubleFieldUpdateRequest()
        update_request.new_value = value
        update_request.absolute_feature_name = self._absolute_feature_name
        update_request.header.CopyFrom(self._fs._feature_set_header)
        self._stub.UpdateFeatureImportance(update_request)
        self._refresh()

    @property
    def monitoring(self):
        return Monitoring(self._stub, self._fs, self)

    @property
    def special_data(self):
        return FeatureSpecialData(self)

    @property
    def marked_for_masking(self):
        return self._internal_feature.marked_for_masking

    @property
    def nested_features(self):
        return {
            feature.name: Feature(
                self._stub,
                self._fs,
                feature,
                self._absolute_feature_name + "." + feature.name,
            )
            for feature in self._internal_feature.nested_features
        }

    def _refresh(self):
        self._fs.refresh()
        feature_name_segments = self._absolute_feature_name.split(".")
        output_feature = [
            f
            for f in self._fs._feature_set.features
            if f.name == feature_name_segments[0]
        ][0]
        for segment in feature_name_segments[1:]:
            output_feature = [
                f for f in output_feature.nested_features if f.name == segment
            ][0]
        self._internal_feature = output_feature

    def __repr__(self):
        return Utils.pretty_print_proto(self._internal_feature)


class FeatureProfile:
    def __init__(self, stub, feature_set, feature):
        self._stub = stub
        self._fs = feature_set
        self._feature = feature

    @property
    def feature_type(self):
        return FeatureSetApi.FeatureTypeInternal.FeatureType.Name(
            self._feature._internal_feature.profile.feature_type
        )

    @feature_type.setter
    def feature_type(self, value):
        valid_values = FeatureSetApi.FeatureType.keys()
        if value.upper() in valid_values:
            update_request = pb.FeatureTypeUpdateRequest()
            update_request.new_value = FeatureSetApi.FeatureType.Value(value.upper())
            update_request.absolute_feature_name = self._feature._absolute_feature_name
            update_request.header.CopyFrom(self._fs._feature_set_header)
            self._stub.UpdateFeatureType(update_request)
            self._feature._refresh()
        else:
            raise Exception(
                "Invalid feature type. Supported values are: "
                + ", ".join(map(str, valid_values))
            )

    @property
    def categorical_statistics(self):
        return CategoricalStatistics(self._feature._internal_feature)

    @property
    def statistics(self):
        return FeatureStatistics(self._feature._internal_feature)

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature._internal_feature.profile)


class FeatureStatistics:
    def __init__(self, feature):
        self._feature = feature
        self._stats = self._feature.categorical

    @property
    def max(self):
        return self._stats.max

    @property
    def mean(self):
        return self._stats.mean

    @property
    def median(self):
        return self._stats.median

    @property
    def min(self):
        return self._stats.min

    @property
    def stddev(self):
        return self._stats.stddev

    @property
    def stddev_rec_count(self):
        return self._stats.stddev_rec_count

    @property
    def null_count(self):
        return self._stats.null_count

    @property
    def nan_count(self):
        return self._stats.nan_count

    @property
    def unique(self):
        return self._stats.unique

    def __repr__(self):
        return Utils.pretty_print_proto(self._stats)


class CategoricalStatistics:
    def __init__(self, feature):
        self._feature = feature
        self._categorical = self._feature.categorical

    @property
    def unique(self):
        return self._categorical.unique

    @property
    def top(self):
        return [FeatureTop(top) for top in self._categorical.top]

    def __repr__(self):
        return Utils.pretty_print_proto(self._categorical)


class FeatureTop:
    def __init__(self, feature):
        self._feature = feature
        self._top = self._feature.top

    @property
    def name(self):
        return self._top.name

    @property
    def count(self):
        return self._top.count

    def __repr__(self):
        return Utils.pretty_print_proto(self._top)


class Monitoring:
    def __init__(self, stub, feature_set, feature):
        self._stub = stub
        self._fs = feature_set
        self._feature = feature

    @property
    def anomaly_detection(self):
        return self._feature._internal_feature.monitoring.anomaly_detection

    @anomaly_detection.setter
    def anomaly_detection(self, value):
        update_request = pb.FeatureBooleanFieldUpdateRequest()
        update_request.new_value = value
        update_request.absolute_feature_name = self._feature._absolute_feature_name
        update_request.header.CopyFrom(self._fs._feature_set_header)
        self._stub.UpdateFeatureAnomalyDetection(update_request)
        self._feature._refresh()

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature._internal_feature.monitoring)


class FeatureSpecialData:
    def __init__(self, feature):
        self._feature = feature

    @property
    def spi(self):
        return self._feature._internal_feature.special_data.spi

    @property
    def pci(self):
        return self._feature._internal_feature.special_data.pci

    @property
    def rpi(self):
        return self._feature._internal_feature.special_data.rpi

    @property
    def demographic(self):
        return self._feature._internal_feature.special_data.demographic

    @property
    def sensitive(self):
        return self._feature._internal_feature.special_data.sensitive

    def __repr__(self):
        return Utils.pretty_print_proto(self._feature._internal_feature.special_data)



import re
import sys
import webbrowser

import grpc
from google.protobuf.empty_pb2 import Empty

import featurestore.core.CoreService_pb2 as pb
from featurestore.core.config import ConfigUtils

from .collections.pats import PersonalAccessTokens


class AuthWrapper:
    ACCESS_TOKEN_EXPIRES_SOON_SECS = 60

    def __init__(self, stub):
        self._stub = stub
        self._access_token = None
        self._access_token_expire = None
        self._get_access_token_external = None
        self._props = ConfigUtils.collect_properties()
        self.pats = PersonalAccessTokens(self._stub)

    def get_active_user(self):
        request = Empty()
        return self._stub.GetActiveUser(request).user

    def set_obtain_access_token_method(self, method):
        self._get_access_token_external = method

    def logout(self):
        is_success, token_or_error = self._obtain_token()
        if is_success:
            if not AuthWrapper._is_personal_access_token(self._props["token"].data):
                request = pb.LogoutRequest()
                request.refresh_token = self._props["token"].data
                self._stub.Logout(request)
                self._access_token = None
                self._access_token_expire = None
            ConfigUtils.delete_property(self._props, ConfigUtils.TOKEN_KEY)
            print("You have been logged out.")
        else:
            print("You are not logged in.")

    def login(self):
        sys.tracebacklimit = None
        for response in self._stub.Login(Empty()):
            if response.HasField("login_url"):
                try:
                    webbrowser.get()
                except webbrowser.Error:
                    print(
                        f"Browser is not supported: Please visit {response.login_url} to continue authentication."
                    )
                print(f"Opening browser to visit: {response.login_url}")
                webbrowser.open(response.login_url)
            elif response.HasField("refresh_token"):
                self.set_auth_token(response.refresh_token)
            else:
                sys.tracebacklimit = 0
                raise AuthException("Incorrect response")

    def set_auth_token(self, token):
        sys.tracebacklimit = None
        ConfigUtils.store_token(self._props, token)
        if not AuthWrapper._is_personal_access_token(token):
            self._access_token = None
            self._access_token_expire = None
            is_success, token_or_error = self._obtain_token()
            if not is_success:
                sys.tracebacklimit = 0
                raise AuthException(token_or_error)

    @staticmethod
    def _is_personal_access_token(token: str) -> bool:
        return re.match(r"^[a-z0-9]{3}_.*", token)

    def _is_access_token_expired(self):
        return self._access_token is None or (
            self._access_token_expire is not None
            and self._access_token_expire <= AuthWrapper.ACCESS_TOKEN_EXPIRES_SOON_SECS
        )

    def _obtain_token(self):
        if self._get_access_token_external is not None:
            return True, self._get_access_token_external()
        elif ConfigUtils.TOKEN_KEY not in self._props:
            return (
                False,
                "You are not authenticated. Set personal access token or execute client.auth.login() method",
            )
        elif AuthWrapper._is_personal_access_token(ConfigUtils.get_token(self._props)):
            return True, ConfigUtils.get_token(self._props)
        elif self._is_access_token_expired():
            request = pb.RefreshTokenRequest()
            request.refresh_token = ConfigUtils.get_token(self._props)
            try:
                resp = self._stub.GetAccessToken(request)
            except grpc.RpcError as e:
                return (
                    False,
                    f"The authentication token is no longer valid. Please login again.",
                )

            self._access_token = resp.access_token
            ConfigUtils.store_token(self._props, resp.refresh_token)
            self._access_token_expire = resp.expires_in
            return True, self._access_token
        else:
            return True, self._access_token

    def __repr__(self):
        return "This class wraps together methods related to Authentication"


class AuthException(Exception):
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: featurestore/core/Events.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='featurestore/core/Events.proto',
  package='featurestore.core',
  syntax='proto3',
  serialized_options=b'\n\030ai.h2o.featurestore.coreB\013EventsProto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1e\x66\x65\x61turestore/core/Events.proto\x12\x11\x66\x65\x61turestore.core\"f\n\x11ModificationEvent\x12H\n\x13\x66\x65\x61ture_set_deleted\x18\x01 \x01(\x0b\x32).featurestore.core.FeatureSetDeletedEventH\x00\x42\x07\n\x05value\"0\n\x16\x46\x65\x61tureSetDeletedEvent\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\tB\'\n\x18\x61i.h2o.featurestore.coreB\x0b\x45ventsProtob\x06proto3'
)




_MODIFICATIONEVENT = _descriptor.Descriptor(
  name='ModificationEvent',
  full_name='featurestore.core.ModificationEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set_deleted', full_name='featurestore.core.ModificationEvent.feature_set_deleted', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='featurestore.core.ModificationEvent.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=53,
  serialized_end=155,
)


_FEATURESETDELETEDEVENT = _descriptor.Descriptor(
  name='FeatureSetDeletedEvent',
  full_name='featurestore.core.FeatureSetDeletedEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set_id', full_name='featurestore.core.FeatureSetDeletedEvent.feature_set_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=157,
  serialized_end=205,
)

_MODIFICATIONEVENT.fields_by_name['feature_set_deleted'].message_type = _FEATURESETDELETEDEVENT
_MODIFICATIONEVENT.oneofs_by_name['value'].fields.append(
  _MODIFICATIONEVENT.fields_by_name['feature_set_deleted'])
_MODIFICATIONEVENT.fields_by_name['feature_set_deleted'].containing_oneof = _MODIFICATIONEVENT.oneofs_by_name['value']
DESCRIPTOR.message_types_by_name['ModificationEvent'] = _MODIFICATIONEVENT
DESCRIPTOR.message_types_by_name['FeatureSetDeletedEvent'] = _FEATURESETDELETEDEVENT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ModificationEvent = _reflection.GeneratedProtocolMessageType('ModificationEvent', (_message.Message,), {
  'DESCRIPTOR' : _MODIFICATIONEVENT,
  '__module__' : 'featurestore.core.Events_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.ModificationEvent)
  })
_sym_db.RegisterMessage(ModificationEvent)

FeatureSetDeletedEvent = _reflection.GeneratedProtocolMessageType('FeatureSetDeletedEvent', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESETDELETEDEVENT,
  '__module__' : 'featurestore.core.Events_pb2'
  # @@protoc_insertion_point(class_scope:featurestore.core.FeatureSetDeletedEvent)
  })
_sym_db.RegisterMessage(FeatureSetDeletedEvent)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import os

from google.protobuf import json_format


class Utils:
    @staticmethod
    def pretty_print_proto(m):
        return json_format.MessageToJson(m, including_default_value_fields=True)

    @staticmethod
    def timestamp_to_string(timestamp):
        if timestamp.ByteSize() != 0:
            return timestamp.ToDatetime().isoformat()
        else:
            return None

    @staticmethod
    def read_env(variable_name, source):
        value = os.environ.get(variable_name)
        if value is None:
            raise Exception(
                "Environment variable "
                + variable_name
                + " is missing, it is required to read from "
                + source
                + " data source."
            )
        else:
            return value
